/*
 * pulseGen: a pulse generator allowing to generate 
 * a sine wave
 * a square wave
 * a sawtooth signal
 * a triangular wave with a frequency of ~ 100 Hz
 * this uses the bcm2835 library which allows setup of the clock divider
 * and thus clock rates on the I2C bus of up to 1.689 MHz in place of the
 * standard 100 kbps rate of standard I2C transfers 
 * copyright U. Raich 2017
 * This program is part of the workshop on embedded systems held at 
 * the University of Cape Coast in the winter semester 2017
 * It has been released under the GNU General Public License
 * see gnu.org/licenses/gpl.html
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <stdbool.h>
#include <unistd.h>
#include <bcm2835.h>

#define MCP4725_ADDRESS  0x62

#define NO_OF_SAMPLES 100
#define RESOLUTION    4095         /* 12 bits */

bool mcp4725Debug;

void mcp4725SetDebug(bool onOff) {
  mcp4725Debug=onOff;
}


int main(int argc, char ** argv) {
  enum waveType {sine,rectangular,triangular,sawtooth,highspeed};
  enum waveType wt;
  char *waveTypeString[]={"sine","rectangular","triangular","sawtooth","highspeed"};
  bool found=false;
  unsigned short samples[NO_OF_SAMPLES];
  unsigned short *samplePtr;
  
  int retCode,i;
  char i2cData[2];
  
  if (argc != 2) {
    printf("Usage: %s wave type\n",argv[0]);
    printf("where \"wave type\" is one of: sine, rectangual, triangulare, sawtooth, highspeed\n");
    exit(-1);
  }
  /*
   * find out which wave type the user has chosen
   */
  for (wt = sine; wt <highspeed+1;wt++)
    if (!strcmp(argv[1],waveTypeString[wt])) {
      found = true;
      break;
    }
  if (!found) {
    printf("pulseGen: wave form %s is not implemented\n",argv[1]);
    exit(-1);
  }

  /*
   * depending on the wave type the user has selected, fill the sample array
   */
  switch(wt) {
  case sine:
    printf("# Generate sine wave\n");
    for (i=0;i<NO_OF_SAMPLES;i++)
      samples[i] = (sin(2*M_PI*i/NO_OF_SAMPLES)+1)*RESOLUTION/2;
    if (mcp4725Debug)
      for (i=0;i<NO_OF_SAMPLES;i++)
	printf("%d\n",samples[i]);
    break;
    
  case rectangular:
    printf("# Generate rectangular wave\n");
    for (i=0;i<NO_OF_SAMPLES;i++)
      samples[i] = (i<NO_OF_SAMPLES/2) ? 0 : RESOLUTION;
    if (mcp4725Debug)
      for (i=0;i<NO_OF_SAMPLES;i++)
	printf("%d\n",samples[i]);      
    break;
    
  case triangular:
    printf("# Generate triangular wave\n");
    for (i=0;i<NO_OF_SAMPLES/2;i++)
      samples[i]=2*(double)RESOLUTION/(double)NO_OF_SAMPLES*i;
    for (i=NO_OF_SAMPLES/2; i< NO_OF_SAMPLES;i++)
      samples[i]=RESOLUTION - 2*(double)RESOLUTION/(double)NO_OF_SAMPLES*(i-NO_OF_SAMPLES/2);
    if (mcp4725Debug)
      for (i=0;i<NO_OF_SAMPLES;i++)
	printf("%d\n",samples[i]);	
    break;
    
  case sawtooth:
    printf("# Generate sawtooth wave\n");
    for (i=0;i<NO_OF_SAMPLES;i++)
      samples[i]=(double)RESOLUTION*i/(double)NO_OF_SAMPLES;
    if (mcp4725Debug)
      for (i=0;i<NO_OF_SAMPLES;i++)
	printf("%d\n",samples[i]);    
    break;
    
  case highspeed:
    for (i=0;i<NO_OF_SAMPLES;i++)
      samples[i]=(i&1) ? 0 : 0xfff;
    if (mcp4725Debug)
      for (i=0;i<NO_OF_SAMPLES;i++)
	printf("%d\n",samples[i]);    
    break;        
  }
  
  if ((retCode=bcm2835_init()) < 0) {
    printf("pulseGen: Could not open bcm2835 library! Are you root?\n");
    return -1;
  }
  else
    printf("pulseGen: bcm2835 library successfully opened!\n");

  /*
    now open the I2C bus
  */
  if ((retCode=bcm2835_i2c_begin()) < 0) {
    printf("pulseGen: error when opening the I2C bus\n");
    return -1;
  }
  else
    printf("pulseGen: I2C bus successfully opened\n");

  bcm2835_i2c_setSlaveAddress(MCP4725_ADDRESS);
  bcm2835_i2c_setClockDivider(BCM2835_I2C_CLOCK_DIVIDER_148);
  printf("pulseGen: slave address set to 0x%02x\n",MCP4725_ADDRESS);
  printf("pulseGen: bit rate set to 1.689 Mbps\n");
  
  /* 
   * now all we need to do is to go through the sample table with a pointer
   * and send the values to the hardware
   */

  if (mcp4725Debug)
    printf("pulseGen: pigpio successfully initialized\n");
   
  printf("pulseGen: Generating the %s wave\n", waveTypeString[wt]);
  for(;;) {
    samplePtr = samples;
    for (i=0;i<NO_OF_SAMPLES;i++) {
      i2cData[0]   = (*samplePtr) >> 8;
      if (mcp4725Debug) {
	i2cData[1] = *samplePtr & 0xff;
	bcm2835_i2c_write(i2cData,2);
	printf("pulseGen: sending 0x%03x reg: 0x%1x value: 0x%02x\n",
	       *samplePtr,i2cData[0],i2cData[1]);
	samplePtr++;
      }
      else {
	i2cData[1] = *samplePtr++ & 0xff;
	bcm2835_i2c_write(i2cData,2);
      }
      usleep(10);     
    } 
  }
  return 0;

  
}
